from langchain.embeddings.openai import OpenAIEmbeddings
from langchain.vectorstores import Chroma
from langchain.text_splitter import CharacterTextSplitter
from langchain import VectorDBQA
from langchain.document_loaders import WebBaseLoader

"""
Notes:

- requires torch etc for embedding. Requires installing the optional `embedding` group of dependencies.
- currently just loads one html page, could do something more like ReadTheDocsLoader
  to load all.

"""

def get_duckdb_docs_chain(llm):
    embeddings = OpenAIEmbeddings()
    loader = WebBaseLoader("https://duckdb.org/docs/sql/introduction")
    docs = loader.load()
    text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
    texts = text_splitter.split_documents(docs)
    docdb = Chroma.from_documents(texts, embeddings, collection_name="duckdb")
    return VectorDBQA.from_chain_type(llm=llm, chain_type="stuff", vectorstore=docdb)




import time
from typing import Any
from rich import print
from langchain.tools import BaseTool

from qabot.duckdb_query import run_sql_catch_error


class DuckDBTool(BaseTool):
    name = "execute"
    description = """useful for when you need to run SQL queries against a DuckDB database.
    Input to this tool is a detailed and correct SQL query, output is a result from the database.
    If the query is not correct, an error message will be returned. 
    If an error is returned, rewrite the query, check the query, and try again.
    """

    database: Any = None

    def __init__(self, engine, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.database = engine

    def _run(self, query: str) -> str:
        #
        query_result = run_sql_catch_error(self.database, query)
        print(f"[pink]{query_result}[/pink]")
        time.sleep(0.2)
        return query_result


    async def _arun(self, query: str) -> str:
        raise NotImplementedError("DuckDBTool does not support async")





from langchain import LLMChain
from langchain.agents import AgentExecutor, Tool, ZeroShotAgent

from qabot.tools.duckdb_execute_tool import DuckDBTool
from qabot.duckdb_query import run_sql_catch_error
from qabot.tools.describe_duckdb_table import describe_table_or_view


def get_duckdb_data_query_chain(llm, database, callback_manager=None, verbose=False):
    tools = [
        Tool(
            name="Show Tables",
            func=lambda _: run_sql_catch_error(database, "show tables;"),
            description="Useful to show the available tables and views. Empty input required."
        ),
        Tool(
            name="Describe Table",
            func=lambda table: describe_table_or_view(database, table),
            description="Useful to show the column names and types of a table or view. Use a valid table name as the input."
        ),
        Tool(
            name="Query Inspector",
            func=lambda query: query.strip('"').strip("'"),
            description="Useful to show the query before execution. Always inspect your query before execution. Input MUST be on one line."
        ),
        DuckDBTool(engine=database),
    ]

    # prompt = PromptTemplate(
    #     input_variables=["input", "agent_scratchpad"],
    #     template=_DEFAULT_TEMPLATE,
    # )

    prompt = ZeroShotAgent.create_prompt(
        tools,
        prefix=prefix,
        suffix=suffix,
        input_variables=["input", "agent_scratchpad", 'table_names'],
    )

    llm_chain = LLMChain(llm=llm, prompt=prompt)

    tool_names = [tool.name for tool in tools]

    agent = ZeroShotAgent(llm_chain=llm_chain, allowed_tools=tool_names,)
    agent_executor = AgentExecutor.from_agent_and_tools(
        agent=agent,
        tools=tools,
        callback_manager=callback_manager,
        verbose=verbose,
    )

    return agent_executor


suffix = """After outputting the Action Input you never output an Observation, that will be provided to you.

List the relevant SQL queries you ran in your final answer.

If a query fails, try fix it, if the database doesn't contain the answer, or returns no results,
output a summary of your actions in your final answer. It is important that you use the exact format:

Final Answer: I have successfully created a view of the data.

Queries should be output on one line and don't use any escape characters. 

Let's go! Remember it is important that you use the exact phrase "Final Answer: " to begin your
final answer.

Question: {input}
Thought: I should describe the most relevant tables in the database to see what columns will be useful.
{agent_scratchpad}"""


prefix = """Given an input question, identify the relevant tables and relevant columns, then create
one single syntactically correct DuckDB query to inspect, then execute, before returning the answer. 
If the input is a valid looking SQL query selecting data or creating a view, execute it directly. 

Even if you know the answer, you MUST show you can get the answer from the database.
Inspect your query before execution.

Refuse to delete any data, or drop tables. You only execute one statement at a time. You may import data.

Example imports:
- CREATE table customers AS SELECT * FROM 'data/records.json';
- CREATE VIEW covid AS SELECT * FROM 's3://covid19-lake/data.csv';

Unless the user specifies in their question a specific number of examples to obtain, limit your
select query to returning 5 results.

Pay attention to use only the column names that you can see in the schema description. Pay attention
to which column is in which table.

You have access to the following tables/views:
{table_names}

You have access to the following tools:
"""



# Other examples

"""

An example final answer:
```
Final Answer: There were 109 male passengers who survived.
The following SQL queries were executed to obtain the result:
- SELECT Sex, Survived FROM titanic limit 5;
- CREATE VIEW male_survivors AS SELECT * FROM titanic WHERE Sex = 'male' AND Survived = 1;
- select count(*) from male_survivors;
```


Examples:?

For example:
 
Input: "Create a names table with an id, name and email column"
Thought: "I need to execute a query to create a table called names, with an id, name and email column"
Action: execute
Action Input: "CREATE TABLE names (id INTEGER, name VARCHAR, email VARCHAR);"
Thought: "I should describe the table to make sure it was created correctly"
Action: Describe Table
Action Input: names
Final Answer: <Summary>


Errors should be returned directly:

Input: "Create a names table with an id, name and email column"
Thought: "I need to execute a query to create a table called names, with an id, name and email column"
Action: execute
Action Input: "CREATE TABLE names (id INTEGER, name VARCHAR, email VARCHAR);"
Final Answer: Error: Catalog Error: Table with name "names" already exists!


For example:
 
Input: "count the number of entries in the "addresses" table that belong to each different city filtering out cities with a count below 50"
Thought: "I need to execute a query to count the number of entries in the "addresses" table that belong to each different city filtering out cities with a count below 50"
Action: execute
Action Input: SELECT city, COUNT(*) FROM addresses GROUP BY city HAVING COUNT(*) >= 50 limit 2;
Thought: 
Final Answer:

"""


from typing import Any, Dict, Optional

from langchain.callbacks import OpenAICallbackHandler
from langchain.schema import AgentAction
from rich import print
from rich.progress import Progress


class QACallback(OpenAICallbackHandler):
    def __init__(self, *args, **kwargs):
        self.progress: Progress = kwargs.pop('progress')
        self.chain_task_ids = []
        self.tool_task_id = None

        super().__init__(*args, **kwargs)

    def on_chain_start(self, serialized, inputs, **kwargs):
        self.chain_task_ids.append(self.progress.add_task(f"on chain start"))

        if isinstance(serialized, dict) and 'name' in serialized:
            self.progress.update(self.chain_task_ids[-1], description=f"[yellow]{serialized['name']}")

        elif 'agent_scratchpad' in inputs and len(inputs['agent_scratchpad']):
            self.progress.update(self.chain_task_ids[-1], description=inputs['agent_scratchpad'])

    def on_tool_start(self, serialized: Dict[str, Any], input_str: str, **kwargs):
        self.tool_task_id = self.progress.add_task(description=f"[yellow]Using tool: {serialized['name']}")

    def on_tool_end(self, output: str, color, observation_prefix, llm_prefix, **kwargs):
        if self.tool_task_id is not None:
            self.progress.remove_task(self.tool_task_id)
            self.tool_task_id = None

    def on_agent_action(
        self, action: AgentAction, color: Optional[str] = None, **kwargs
    ):
        """Run on agent requesting an action."""
        print(f"[{color}]{action.log}[/{color}]")

    def on_chain_end(self, outputs, **kwargs):
        super().on_chain_end(outputs, **kwargs)
        if isinstance(outputs, dict) and 'text' in outputs:
            outputs = outputs['text']
            #print(f"[cyan]{outputs}")

        self.progress.update(self.chain_task_ids[-1], description=f"[yellow]Total tokens: {self.total_tokens} {outputs}")
        self.progress.remove_task(self.chain_task_ids.pop())

    def on_agent_finish(
        self, finish, color: Optional[str] = None, **kwargs
    ) -> None:
        """Run on agent end."""
        if 'output' in finish.return_values:
            print(f"[{color}]{finish.return_values['output']}[/{color}]")

    def on_llm_end(self, response, **kwargs):
        super().on_llm_end(response, **kwargs)
        # Could add tokens to database, or progress bar here

        self.progress.update(self.chain_task_ids[-1], description=f"[yellow]Total Tokens {self.total_tokens}")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['qabot', 'qabot.agents', 'qabot.tools']

package_data = \
{'': ['*']}

install_requires = \
['duckdb-engine>=0.6.9,<0.7.0',
 'langchain>=0.0.107,<0.0.108',
 'openai>=0.27.0,<0.28.0',
 'psycopg2-binary>=2.9.5,<3.0.0',
 'rich>=13.3.1,<14.0.0',
 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['qabot = qabot.cli:run']}

setup_kwargs = {
    'name': 'qabot',
    'version': '0.2.7',
    'description': 'Query local or remote data files with natural language queries powered by OpenAI and DuckDB.',
    'long_description': '# qabot\n\nQuery local or remote files with natural language queries powered by\n`langchain` and `gpt-3.5-turbo` and `duckdb` 🦆.\n\n\n\nWorks on local CSV files:\n\n![](.github/local_csv_query.png)\n\nremote CSV files:\n\n```\n$ qabot \\\n    -f https://www.stats.govt.nz/assets/Uploads/Environmental-economic-accounts/Environmental-economic-accounts-data-to-2020/renewable-energy-stock-account-2007-2020-csv.csv \\\n    -q "How many Gigawatt hours of generation was there for Solar resources in 2015 through to 2020?"\n```\n\n\nEven on (public) data stored in S3:\n\n![](.github/external_s3_data.png)\n\nYou can even load data from disk via the natural language query, but that doesn\'t always work...\n\n\n> "Load the file \'data/titanic_survival.parquet\' into a table called \'raw_passengers\'. Create a view of the raw passengers table for just the male passengers. What was the average fare for surviving male passengers?"\n\n\nAfter a bit of back and forth with the model, it gets there:\n\n> The average fare for surviving male passengers from the \'male_passengers\' view where the passenger survived is 40.82. I ran the query: SELECT AVG(Fare) FROM male_passengers WHERE Survived = 1 AND Sex = \'male\';\nThe average fare for surviving male passengers is 40.82.\n\n\n## Quickstart\n\nYou need to set the `OPENAI_API_KEY` environment variable to your OpenAI API key, \nwhich you can get from [here](https://platform.openai.com/account/api-keys).\n\nInstall the `qabot` command line tool using pip/poetry:\n\n\n```bash\n$ pip install qabot\n```\n\nThen run the `qabot` command with either local files (`-f my-file.csv`) or a database connection string.\n\nNote if you want to use a database, you will need to install the relevant drivers, \ne.g. `pip install psycopg2-binary` for postgres.\n\n\n\n## Examples\n\n### Local CSV file/s\n\n```bash\n$ qabot -q "how many passengers survived by gender?" -f data/titanic.csv\n🦆 Loading data from files...\nLoading data/titanic.csv into table titanic...\n\nQuery: how many passengers survived by gender?\nResult:\nThere were 233 female passengers and 109 male passengers who survived.\n\n\n 🚀 any further questions? [y/n] (y): y\n\n 🚀 Query: what was the largest family who did not survive? \nQuery: what was the largest family who did not survive?\nResult:\nThe largest family who did not survive was the Sage family, with 8 members.\n\n 🚀 any further questions? [y/n] (y): n\n```\n\n\n## Intermediate steps and database queries\n\nUse the `-v` flag to see the intermediate steps and database queries.\n\nSometimes it takes a long route to get to the answer, but it\'s interesting to see how it gets there:\n\n\n```\nqabot -f data/titanic.csv -q "how many passengers survived by gender?" -v\n🦆 Loading data from files...\nQuery: how many passengers survived by gender?\nI need to check the columns in the \'titanic\' table to see which ones contain gender and survival information.\nAction: Describe Table\nAction Input: titanic\n\nObservation: titanic\n\n┌─────────────┬─────────────┬─────────┬─────────┬─────────┬───────┐\n│ column_name │ column_type │  null   │   key   │ default │ extra │\n│   varchar   │   varchar   │ varchar │ varchar │ varchar │ int32 │\n├─────────────┼─────────────┼─────────┼─────────┼─────────┼───────┤\n│ PassengerId │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n│ Survived    │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n│ Pclass      │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n│ Name        │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n│ Sex         │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n│ Age         │ DOUBLE      │ YES     │ NULL    │ NULL    │  NULL │\n│ SibSp       │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n│ Parch       │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n│ Ticket      │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n│ Fare        │ DOUBLE      │ YES     │ NULL    │ NULL    │  NULL │\n│ Cabin       │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n│ Embarked    │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n├─────────────┴─────────────┴─────────┴─────────┴─────────┴───────┤\n│ 12 rows                                               6 columns │\n└─────────────────────────────────────────────────────────────────┘\n\nI need to create a view that only includes the columns I need for this question.\nAction: Data Op\nAction Input: \n        CREATE VIEW titanic_gender_survival AS\n        SELECT Sex, Survived\n        FROM titanic\nThought:\n\n> Entering new AgentExecutor chain...\nThis is a valid SQL query creating a view. We can execute it directly.\nAction: execute\nAction Input: \n        CREATE VIEW titanic_gender_survival AS\n        SELECT Sex, Survived\n        FROM titanic\nObservation: No output\nThought:The view has been created successfully. We can now query it.\nAction: execute\nAction Input: SELECT * FROM titanic_gender_survival LIMIT 5\nObservation: \n┌─────────┬──────────┐\n│   Sex   │ Survived │\n│ varchar │  int64   │\n├─────────┼──────────┤\n│ male    │        0 │\n│ female  │        1 │\n│ female  │        1 │\n│ female  │        1 │\n│ male    │        0 │\n└─────────┴──────────┘\n\nThought:The view has been created successfully and we can see the first 5 rows of the view. The final answer is the first 5 rows of the titanic_gender_survival view, showing the sex and survival status of passengers on the \nTitanic.\nFinal Answer: \n┌─────────┬──────────┐\n│   Sex   │ Survived │\n│ varchar │  int64   │\n├─────────┼──────────┤\n│ male    │        0 │\n│ female  │        1 │\n│ female  │        1 │\n│ female  │        1 │\n│ male    │        0 │\n└─────────┴──────────┘\n\n> Finished chain.\n\nObservation: ┌─────────┬──────────┐\n│   Sex   │ Survived │\n│ varchar │  int64   │\n├─────────┼──────────┤\n│ male    │        0 │\n│ female  │        1 │\n│ female  │        1 │\n│ female  │        1 │\n│ male    │        0 │\n└─────────┴──────────┘\nI need to group the data by gender and count the number of survivors for each group.\nAction: Data Op\nAction Input: \n        SELECT Sex, SUM(Survived) AS num_survived\n        FROM titanic_gender_survival\n        GROUP BY Sex\nThought:\n\n> Entering new AgentExecutor chain...\nWe need to check if the table titanic_gender_survival exists and if it has the columns Sex and Survived.\nAction: Describe Table\nAction Input: titanic_gender_survival\nObservation: titanic_gender_survival\n\n┌─────────────┬─────────────┬─────────┬─────────┬─────────┬───────┐\n│ column_name │ column_type │  null   │   key   │ default │ extra │\n│   varchar   │   varchar   │ varchar │ varchar │ varchar │ int32 │\n├─────────────┼─────────────┼─────────┼─────────┼─────────┼───────┤\n│ Sex         │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n│ Survived    │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n└─────────────┴─────────────┴─────────┴─────────┴─────────┴───────┘\n\nThought:The table titanic_gender_survival exists and has the columns Sex and Survived. We can now run the query.\nAction: execute\nAction Input: \n```\nSELECT Sex, SUM(Survived) AS num_survived\nFROM titanic_gender_survival\nGROUP BY Sex\nLIMIT 5\n```\n\nObservation: \n┌─────────┬──────────────┐\n│   Sex   │ num_survived │\n│ varchar │    int128    │\n├─────────┼──────────────┤\n│ male    │          109 │\n│ female  │          233 │\n└─────────┴──────────────┘\n\nThought:The query returned the number of survivors grouped by gender. The table titanic_gender_survival has been used. \nFinal Answer: The number of survivors grouped by gender are: \n- 109 males survived\n- 233 females survived.\n\n> Finished chain.\n\nObservation: The number of survivors grouped by gender are: \n- 109 males survived\n- 233 females survived.\nIntermediate Steps: \n  Step 1\n\n    Describe Table\n      titanic\n\n      titanic\n\n    ┌─────────────┬─────────────┬─────────┬─────────┬─────────┬───────┐\n    │ column_name │ column_type │  null   │   key   │ default │ extra │\n    │   varchar   │   varchar   │ varchar │ varchar │ varchar │ int32 │\n    ├─────────────┼─────────────┼─────────┼─────────┼─────────┼───────┤\n    │ PassengerId │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n    │ Survived    │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n    │ Pclass      │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n    │ Name        │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n    │ Sex         │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n    │ Age         │ DOUBLE      │ YES     │ NULL    │ NULL    │  NULL │\n    │ SibSp       │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n    │ Parch       │ BIGINT      │ YES     │ NULL    │ NULL    │  NULL │\n    │ Ticket      │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n    │ Fare        │ DOUBLE      │ YES     │ NULL    │ NULL    │  NULL │\n    │ Cabin       │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n    │ Embarked    │ VARCHAR     │ YES     │ NULL    │ NULL    │  NULL │\n    ├─────────────┴─────────────┴─────────┴─────────┴─────────┴───────┤\n    │ 12 rows                                               6 columns │\n    └─────────────────────────────────────────────────────────────────┘\n\n    \n\n  Step 2\n\n    Data Op\n      CREATE VIEW titanic_gender_survival AS\n            SELECT Sex, Survived\n            FROM titanic\n\n      ┌─────────┬──────────┐\n    │   Sex   │ Survived │\n    │ varchar │  int64   │\n    ├─────────┼──────────┤\n    │ male    │        0 │\n    │ female  │        1 │\n    │ female  │        1 │\n    │ female  │        1 │\n    │ male    │        0 │\n    └─────────┴──────────┘\n\n    \n\n  Step 3\n\n    Data Op\n      SELECT Sex, SUM(Survived) AS num_survived\n            FROM titanic_gender_survival\n            GROUP BY Sex\n\n      The number of survivors grouped by gender are: \n    - 109 males survived\n    - 233 females survived.\n\n    \n\n\nThought:\n\n\nResult:\n109 males and 233 females survived.\n```\n\n## Data accessed via http/s3\n\nUse the `-f <url>` flag to load data from a url, e.g. a csv file on s3:\n\n```bash\n$ qabot -f s3://covid19-lake/enigma-jhu-timeseries/csv/jhu_csse_covid_19_timeseries_merged.csv -q "how many confirmed cases are there" -v\n🦆 Loading data from files...\ncreate table jhu_csse_covid_19_timeseries_merged as select * from \'s3://covid19-lake/enigma-jhu-timeseries/csv/jhu_csse_covid_19_timeseries_merged.csv\';\n\nResult:\n264308334 confirmed cases\n```\n\n## Links\n\n- [Python library docs](https://langchain.readthedocs.io)\n- [Agent docs to talk to arbitrary apis via OpenAPI/Swagger](https://langchain.readthedocs.io/en/latest/modules/agents/agent_toolkits/openapi.html)\n- [Agents/Tools to talk SQL](https://langchain.readthedocs.io/en/latest/modules/agents/agent_toolkits/sql_database.html)\n- [Typescript library](https://hwchase17.github.io/langchainjs/docs/overview/)\n\n\n\n## Ideas\n\n- Upgrade to use langchain chat interface\n- Use memory, perhaps wait for langchain\'s next release\n- Decent Python Library API so can be used from other Python code\n- streaming mode to output results as they come in\n- token usage reporting and limits\n- Supervisor agent - assess whether a query is "safe" to run, could ask for user confirmation to run anything that gets flagged.\n- Often we can zero-shot the question and get a single query out - perhaps we try this before the MKL chain\n- test each zeroshot agent individually\n- Pass in the available tables from the first prompt\n- Generate and pass back assumptions made to the user\n- Add an optional "clarify" tool to the chain that asks the user to clarify the question\n- Create a query checker tool that checks if the query looks valid\n- Perhaps an explain query tool that shows the steps taken to get the answer\n- Store all queries, actions, and answers in a table\n- Optional settings to switch to different LLM\n- Inject AWS credentials into duckdb so we can access private resources in S3\n- A version that uses document embeddings - probably not in this app as needs Torch',
    'author': 'Brian Thorne',
    'author_email': 'brian@hardbyte.nz',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import os
from typing import Tuple
from urllib.parse import urlparse
import duckdb
from duckdb import ParserException, ProgrammingError


def uri_validator(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc])
    except:
        return False


def create_duckdb(duckdb_path: str = ':memory:') -> duckdb.DuckDBPyConnection:
    # By default, duckdb is fully in-memory - we can provide a path to get
    # persistent storage

    duckdb_connection = duckdb.connect(duckdb_path)
    try:
        duckdb_connection.sql("INSTALL httpfs;")
        duckdb_connection.sql("LOAD httpfs;")
    except Exception:
        print("Failed to install httpfs extension. Only loading from local files will be supported")

    duckdb_connection.sql("create temporary table if not exists _qabot_queries(query VARCHAR PRIMARY KEY, result VARCHAR)")

    return duckdb_connection


def import_into_duckdb_from_files(duckdb_connection: duckdb.DuckDBPyConnection, files: list[str]) -> Tuple[duckdb.DuckDBPyConnection, list[str]]:

    executed_sql = []
    for i, file_path in enumerate(files, 1):

        executed_sql.append(load_external_data_into_db(duckdb_connection, file_path))

        # Alternative is to allow user to pass in column types:
        # conn.sql(f"""
        # create table {table_name} as (
        #     select * from read_csv_auto(
        #         '%s',
        #         delim='|',
        #         header=True,
        #         --types={'phone': 'VARCHAR'}
        #     )
        # )
        # """ % file_path)

    return duckdb_connection, executed_sql


def load_external_data_into_db(conn: duckdb.DuckDBPyConnection, file_path, allow_view=True):
    # Work out if the filepath is actually a url (e.g. s3://)
    is_url = uri_validator(file_path)
    # Get the file name without extension from the file_path
    table_name, extension = os.path.splitext(os.path.basename(file_path))
    # If the table_name isn't a valid SQL identifier, we'll need to use something else

    table_name = table_name.replace("-", "_").replace(".", "_").replace(" ", "_").replace('/', '_')

    try:
        conn.sql(f"create table t_{table_name} as select 1;")
        conn.sql(f"drop table t_{table_name};")
    except (ParserException, ProgrammingError) as e:
        table_name = "data"

    # The SQLAgent doesn't appear to see view's just yet, so we'll create a table instead
    use_view = allow_view and is_url

    create_statement = f"create {'view' if use_view else 'table'} '{table_name}' as select * from '{file_path}';"

    conn.sql(create_statement)

    return create_statement



from qabot.duckdb_query import run_sql_catch_error


def describe_table_or_view(database, table):
    statement = f"select column_name, data_type from information_schema.columns where table_name='{table}';"
    table_description = run_sql_catch_error(database, statement)
    return f"{table}\n{table_description}"



import textwrap
from typing import List, Optional
import warnings

import typer
from langchain.callbacks.openai_info import OpenAICallbackHandler
from langchain.schema import AgentAction
from rich import print
from langchain.callbacks import get_callback_manager

from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.prompt import Confirm, Prompt

from qabot.caching import configure_caching
from qabot.config import Settings
from qabot.duckdb_manual_data_loader import import_into_duckdb_from_files, create_duckdb
from qabot.agents.agent import create_agent_executor
from qabot.duckdb_query import run_sql_catch_error
from qabot.progress_callback import QACallback

warnings.filterwarnings("ignore")

INITIAL_NON_INTERACTIVE_PROMPT = "🚀 How can I help you explore your database?"
INITIAL_INTERACTIVE_PROMPT = "[bold green] 🚀 How can I help you explore your database?"
FOLLOW_UP_PROMPT = "[bold green] 🚀 any further questions?"
PROMPT = "[bold green] 🚀 Query"

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    pretty_exceptions_enable=True
)


def format_intermediate_steps(intermediate_steps):
    if isinstance(intermediate_steps, list):
        return "\n".join(intermediate_steps)
    else:
        return str(intermediate_steps)


def format_agent_action(agent_action: AgentAction, observation) -> str:
    """
    Sometimes observation is a string, sometimes it is a dict. This function handles both cases.


    """
    result = ''
    internal_result = str(observation).strip()
    logs = ''

    if isinstance(observation, dict):
        if 'input' in observation:
            # should be the same as agent_action.tool_input
            pass
        if 'output' in observation:
            internal_result = observation['output']
        #if 'intermediate_steps' in observation:
        #     observation = format_intermediate_steps(observation['intermediate_steps'])

    if len(agent_action) > 3:
        logs = '\n'.join([textwrap.indent(str(o).strip(), ' '*6) for o in agent_action])

    return f"""
[red]{agent_action.tool.strip()}[/red]
  [green]{agent_action.tool_input.strip()}[/green]

  [blue]{internal_result}[/blue]
  
    [cyan]{str(logs).strip()}[/cyan]

[bold red]{result}[/bold red]
"""


@app.command()
def main(
        query: str = typer.Option("Describe the tables", '-q', '--query', prompt=INITIAL_NON_INTERACTIVE_PROMPT),
        file: Optional[List[str]] = typer.Option(None, "-f", "--file", help="File or url containing data to query"),
        #database_uri: Optional[str] = typer.Option(None, "-d", "--database", help="Database URI (e.g. sqlite:///mydb.db)"),
        table: Optional[List[str]] = typer.Option(None, "--table", "-t", help="Limit queries to these tables (can be specified multiple times)"),
        disable_cache: bool = typer.Option(False, "--disable-cache", help="Disable caching of LLM queries"),
        verbose: bool = typer.Option(False, "-v", "--verbose", help='Essentially debug output'),
):
    """
    Query a database using a simple english query.

    Example:
        qabot -q "What is the average age of the people in the table?"
    """

    settings = Settings()
    executed_sql = ''
    # If files are given load data into local DuckDB
    database_engine = create_duckdb()

    if len(file) > 0:
        if isinstance(file, str):
            file = [file]
        print("[red]🦆[/red] [bold]Loading data from files...[/bold]")
        database_engine, executed_sql = import_into_duckdb_from_files(database_engine, file)
        executed_sql = '\n'.join(executed_sql)
    else:
        print("[red]🦆[/red]")

    with Progress(
        SpinnerColumn(),
        TextColumn("[green][progress.description]{task.description}"),
        transient=False,
    ) as progress:

        output_callback = QACallback(progress=progress)
        openai_callback = OpenAICallbackHandler()
        callback_manager = get_callback_manager()

        callback_manager.add_handler(openai_callback)
        callback_manager.add_handler(output_callback)



        if not disable_cache:
            t = progress.add_task(description="Setting up cache...", total=None)
            configure_caching(settings.QABOT_CACHE_DATABASE_URI)
            progress.remove_task(t)

        t2 = progress.add_task(description="Creating LLM agent using langchain...", total=None)

        agent = create_agent_executor(
            #database_uri=database_uri or settings.QABOT_DATABASE_URI,
            database_engine=database_engine,
            tables=table,
            return_intermediate_steps=True,
            callback_manager=callback_manager,
            verbose=False,
        )

        progress.remove_task(t2)
        chat_history = [f"""
        Startup SQL Queries:
        ```
        {executed_sql}
        ```
        """]

        while True:

            t = progress.add_task(description="Processing query...", total=None)
            print("[bold red]Query: [/][green]" + query)

            inputs = {
                "input": chat_history[0] + query,
                "table_names": run_sql_catch_error(database_engine, "show tables")
                #'chat_history': '\n\n'.join(chat_history)
            }

            result = agent(inputs)

            progress.remove_task(t)

            # Show intermediate steps
            if verbose:
                progress.console.print("[bold red]Intermediate Steps: [/]")
                for i, (agent_action, action_input) in enumerate(result['intermediate_steps'], 1):
                    print(f"  [bold red]Step {i}[/]")
                    print(textwrap.indent(format_agent_action(agent_action, action_input), "    "))

                print()
            print("Total tokens", output_callback.total_tokens)

            # Stop the progress before outputting result and prompting for input
            progress.stop()
            print()

            print("[bold red]Result:[/]\n[bold blue]" + result['output'] + "\n")
            chat_history.append(result['output'])

            if not Confirm.ask(FOLLOW_UP_PROMPT, default=True):
                break

            print()
            query = Prompt.ask(PROMPT)

            if query == "exit" and Confirm.ask("Are you sure you want to Quit?"):
                break

            progress.start()


def run():
    app()


if __name__ == '__main__':
    run()